@@ -24,6 +24,7 @@ from __future__ import absolute_import
 # noteheads. So I think it should be given the created NoteHeadEngravers
 # and then shift the notes.
 
+import copy
 import re
 import weakref
 
@@ -244,6 +245,15 @@ class Voice(HasParent):
         # The timelen of the Voice
         self.m_length = Rat(0, 1)
         self.m_tdict = {}
+    def copy(self, parent):
+        """
+        Return a copy of this Voice object. We make a copy of the dict and
+        the m_length variable, but the dict revers to the same object.
+        """
+        ret = Voice(parent)
+        ret.m_length = Rat(self.m_length.m_num, self.m_length.m_den)
+        ret.m_tdict = self.m_tdict.copy()
+        return ret
     class BarFullException(Exception):
         """
         Exception raised if we try to add a note or rest that is longer than
@@ -286,8 +296,7 @@ class Voice(HasParent):
             # IndexError means that self.m_length is after the last bar, this means
             # that the current bar is full.
             bar = self.w_score().add_bar(None)
-        if (self.m_length + elem[0].m_duration.get_rat_value()
-            <= bar.m_timepos + bar.m_timesig.as_rat()):
+        if (self.m_length + elem[0].m_duration.get_rat_value() <= bar.end()):
             if isinstance(elem[0], Note):
                 # We don't have to set w_parent for the Stem and the notes
                 # here, because the Stem constructor does it.
@@ -364,13 +373,42 @@ class Voice(HasParent):
             max_free = bp.get_free_time()
             #flytt til bar-class
             delta = elem.m_duration.get_rat_value() - self.m_tdict[timepos]['elem'][0].m_duration.get_rat_value()
+            def fix_ties(prev_timepos, cur_timepos, next_timepos):
+                prev_e = self.m_tdict[prev_timepos]['elem'][0]
+                cur_e = self.m_tdict[cur_timepos]['elem'][0]
+                next_e = self.m_tdict[next_timepos]['elem'][0]
+                if isinstance(elem, Note):
+                    t = None
+                    if prev_timepos and isinstance(prev_e, Note) and prev_e.m_tieinfo in ('go', 'start'):
+                        t = 'end'
+                    if isinstance(next_e, Note) and next_e.m_tieinfo in ('go', 'end'):
+                        if t == 'end':
+                            t = 'go'
+                        else:
+                            t = 'start'
+                    cur_e.m_tieinfo = t
+                else:
+                    assert isinstance(elem, Rest)
+                    if prev_timepos and isinstance(prev_e, Note):
+                        if prev_e.m_tieinfo == 'go':
+                            prev_e.m_tieinfo = 'end'
+                        elif prev_e.m_tieinfo == 'start':
+                            prev_e.m_tieinfo = None
+                    if isinstance(next_e, Note):
+                        if next_e.m_tieinfo == 'go':
+                            next_e.m_tieinfo = 'start'
+                        elif next_e.m_tieinfo == 'end':
+                            next_e.m_tieinfo = None
+
             if insert_mode:
                 if max_free >= elem.m_duration.get_rat_value():
-                    self.m_tdict[timepos + Rat(1, 1000000)] = self.m_tdict[timepos]
+                    tmp_timepos = timepos + Rat(1, 1000000)
+                    self.m_tdict[tmp_timepos] = self.m_tdict[timepos]
                     del self.m_tdict[timepos]
                     bp.remove_trailing(elem.m_duration.get_rat_value())
                     stem = Stem(self, [elem], const.UP)
                     self.set_elem(stem, timepos)
+                    fix_ties(self.get_prev_timepos(timepos), timepos, tmp_timepos)
                     bp.remove_skips()
                     bp.repack()
                     bp.fill_skips()
@@ -384,6 +422,7 @@ class Voice(HasParent):
                         bp.remove_trailing(delta)
                     stem = Stem(self, [elem], const.UP)
                     self.set_elem(stem, timepos)
+                    fix_ties(self.get_prev_timepos(timepos), timepos, self.get_next_timepos(timepos))
                     bp.remove_skips()
                     bp.repack()
                     bp.fill_skips()
@@ -408,10 +447,10 @@ class Voice(HasParent):
             start_i = 0
             while start_i < len(v) and v[start_i] < bar.m_timepos:
                 start_i += 1
-        if start_i >= len(v) or v[start_i] >= bar.m_timepos + bar.m_timesig.as_rat():
+        if start_i >= len(v) or v[start_i] >= bar.end():
             start_i = None
         try:
-            end_i = v.index(bar.m_timepos + bar.m_timesig.as_rat())
+            end_i = v.index(bar.end())
         except ValueError:
             end_i = None
         if start_i == None:
@@ -767,9 +806,17 @@ class Bar(object):
                 and voice.m_tdict[timeposes[-1]]['elem'][0].m_tieinfo in ('go', 'start')):
                 voice.untie_next(timeposes[-1])
     def __repr__(self):
-        return "<Bar %i/%i %s at %s>" % (self.m_timesig.m_num,
+        return "<%s %i/%i %s at %s>" % (str(self.__class__).split(".")[-1][:-2], self.m_timesig.m_num,
             self.m_timesig.m_den, self.m_timepos, hex(id(self)))
 
+class PartialBar(Bar):
+    def __init__(self, duration, timesig, timepos):
+        Bar.__init__(self, timesig, timepos)
+        assert isinstance(duration, Duration)
+        self.m_duration = duration
+    def end(self):
+        return self.m_timepos + self.m_duration.get_rat_value()
+
 class BarProxy(object):
     def __init__(self, voice, timepos):
         self.m_voice = voice
@@ -788,7 +835,15 @@ class _StaffCommon(HasParent):
         HasParent.__init__(self, parent)
         self.m_voices = []
         self.add_voice()
+        # I think the only things stored in Staff.m_tdict are "clef" and
+        # "keysig". We don't store time signature changes where, since
+        # Score.m_bars take care about that.
         self.m_tdict = {}
+    def copy(self, parent):
+        staff = self.__class__(parent)
+        staff.m_voices = [v.copy(staff) for v in self.m_voices]
+        staff.m_tdict = self.m_tdict.copy()
+        return staff
     def add_voice(self):
         self.m_voices.append(Voice(self))
         self.w_score().create_shortcuts()
@@ -885,9 +940,18 @@ class RhythmStaff(_StaffCommon):
 
 
 class Score(object):
+    class ConcatException(Exception): pass
+    class StaffCountException(ConcatException): pass
+    class StaffTypeException(ConcatException): pass
+    class VoiceCountException(ConcatException): pass
     def __init__(self):
         self.m_staffs = []
         self.m_bars = []
+    def copy(self):
+        score = Score()
+        score.m_staffs = [s.copy(score) for s in self.m_staffs]
+        score.m_bars = copy.deepcopy(self.m_bars)
+        return score
     def add_staff(self, staff_class=Staff):
         self.m_staffs.append(staff_class(self))
         self.create_shortcuts()
@@ -901,21 +965,38 @@ class Score(object):
             setattr(self, "staff%i" % (staff_idx + 1), staff)
             for voice_idx, voice in enumerate(staff.m_voices):
                 setattr(self, "voice%i%i" % (staff_idx + 1, voice_idx + 1), voice)
+    def _get_new_bar_timepos(self):
+        """
+        Return the timepos where the next bar will be added.
+        """
+        if self.m_bars:
+            return self.m_bars[-1].end()
+        return Rat(0, 1)
+    def _get_new_bar_timesig(self, timesig):
+        """
+        Return the time signature the next bar will get.
+        """
+        if timesig:
+            return timesig
+        elif self.m_bars:
+            return self.m_bars[-1].m_timesig
+        else:
+            return TimeSignature(4, 4)
     def add_bar(self, timesig):
         """
         If timesig is None, then we use the same timesig as the last bar.
         Return the added bar.
         """
-        if self.m_bars:
-            c = self.m_bars[-1].m_timepos + self.m_bars[-1].m_timesig.as_rat()
-        else:
-            c = Rat(0, 1)
-        if timesig:
-            self.m_bars.append(Bar(timesig, c))
-        elif self.m_bars:
-            self.m_bars.append(Bar(self.m_bars[-1].m_timesig, c))
-        else:
-            self.m_bars.append(Bar(TimeSignature(4, 4), c))
+        self.m_bars.append(Bar(
+            self._get_new_bar_timesig(timesig), self._get_new_bar_timepos()))
+        return self.m_bars[-1]
+    def add_partial_bar(self, duration, timesig):
+        """
+        Set to the duration of the pickup bar if we want one.
+        This must be called before bars are added with Score.add_bar
+        """
+        self.m_bars.append(PartialBar(duration,
+            self._get_new_bar_timesig(timesig), self._get_new_bar_timepos()))
         return self.m_bars[-1]
     def get_bar_at(self, timepos):
         """
@@ -923,7 +1004,7 @@ class Score(object):
         is after the last bar.
         """
         for idx, bar in enumerate(self.m_bars):
-            if bar.m_timepos <= timepos <bar.m_timepos + bar.m_timesig.as_rat():
+            if bar.m_timepos <= timepos < bar.end():
                 return bar
             if bar.m_timepos > timepos:
                 return IndexError(timepos)
@@ -975,4 +1056,72 @@ class Score(object):
                     retval.append([True, timepos - ppos])
                     ppos = timepos
         return retval
-
+    @staticmethod
+    def concat(s1, s2):
+        """
+        Concatenate the two scores, and return a new score. Both scores
+        need to have the exact same staff and voice layout.
+        """
+        assert isinstance(s1, Score)
+        assert isinstance(s2, Score)
+        if len(s1.m_staffs) != len(s2.m_staffs):
+            raise Score.StaffCountException()
+        if [type(x) for x in s1.m_staffs] != [type(x) for x in s2.m_staffs]:
+            raise Score.StaffTypeException()
+        for idx, (st1, st2) in enumerate(zip(s1.m_staffs, s2.m_staffs)):
+            if len(st1.m_voices) != len(st2.m_voices):
+                raise Score.VoiceCountException()
+        ret = s1.copy()
+        if not s1.m_staffs:
+            return s1
+        # do the adding
+        for bar in s2.m_bars:
+            bar.m_timepos = ret.m_bars[-1].end()
+            ret.m_bars.append(bar)
+        ret.create_shortcuts()
+        if s1.m_bars:
+            start = s1.m_bars[-1].end()
+        else:
+            start = Rat(0, 1)
+        s2.create_shortcuts() # FIXME why?
+        for staff_idx, staff in enumerate(s1.m_staffs):
+            for voice_idx, voice in enumerate(staff.m_voices):
+                for k in s2.m_staffs[staff_idx].m_voices[voice_idx].m_tdict:
+                    ret.m_staffs[staff_idx].m_voices[voice_idx].m_tdict[k + start] = s2.m_staffs[staff_idx].m_voices[voice_idx].m_tdict[k]
+        return ret
+    @staticmethod
+    def concat2(s1, s2):
+        """
+        Return a new Score object concatenating the two scores. This is
+        intended return value is intended for playback only, since the
+        staffs placed below each other. So the first score will have empty
+        bars at the end, and the last score will have empty bars at the
+        beginning.
+        """
+        assert isinstance(s1, Score)
+        assert isinstance(s2, Score)
+        ret = s1.copy()
+        if s1.m_bars:
+            start = s1.m_bars[-1].end()
+        else:
+            start = Rat(0, 1)
+        for bar in s2.m_bars:
+            ret.m_bars.append(Bar(bar.m_timesig, ret.m_bars[-1].end()))
+        for staff_idx, staff in enumerate(s2.m_staffs):
+            ret.add_staff(staff_class=staff.__class__)
+            for k in staff.m_tdict:
+                ret.m_staffs[-1].m_tdict[start + k] = staff.m_tdict[k]
+            for voice_idx, voice in enumerate(staff.m_voices):
+                if voice_idx != 0:
+                    ret.m_staffs[-1].add_voice()
+                # This line make the music from sc2 continue after the
+                # point where the music from sc1 ends.
+                ret.m_staffs[-1].m_voices[-1].m_length = s1.m_bars[-1].end()
+                for elem in s2.m_staffs[staff_idx].m_voices[voice_idx]:
+                    ret.m_staffs[-1].m_voices[-1].append(elem['elem'])
+        ret.create_shortcuts()
+        return ret
+    def __deepcopy__(self, memo):
+        ret = Score()
+        ret.m_bars = copy.deepcopy(self.m_bars)
+        return ret
